﻿Imports System.IO
Imports System.Runtime.InteropServices
Imports System.Text

Module Module_PSN

    ' ----------------------------------------------------- flags
    Const NO_CRC As Int32 = 1
    Const NO_MINMAX As Int32 = 2


    ' ==========================================================================================
    '  PSN4 FILE STRUCTURES
    ' ==========================================================================================

    ' ----------------------------------------------------- PSN4 FILE
    Structure PSN4_File
        Dim Name As String
        Dim VolumeHeader As PSN4_VolumeType1
        Dim Records() As PSN4_Record
    End Structure

    ' ----------------------------------------------------- PSN4 RECORD
    Structure PSN4_Record
        Dim FixedHeader As PSN4_FixedHeader
        Dim FixedHeaderBytes As Byte()       'temporary array of FixedHeaderBytes (used for file read-write and CRC)
        Dim VarHeaders() As PSN4_VarHeader
        Dim NumDataBytes As Int32
        Dim DataInt16() As Int16
        Dim DataInt32() As Int32
        Dim DataSingle() As Single
        Dim DataDouble() As Double
        Dim CRC As UInt16
    End Structure


    ' ==========================================================================================
    '  PSN4 STRUCTURES  ( http://www.seismicnet.com/psnformat4.html )
    ' ==========================================================================================

    ' ----------------------------------------------------- volume header ( 12 bytes )
    <StructLayout _
    (LayoutKind.Sequential, Pack:=1, CharSet:=CharSet.Ansi)> _
    Structure PSN4_VolumeType1
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=10)> Dim volumeIdAndVersion As Char()
        Dim NumberEventRecords As UInt16
    End Structure


    ' ----------------------------------------------------- fixed header ( 154 bytes ) 
    <StructLayout _
    (LayoutKind.Sequential, Pack:=1, CharSet:=CharSet.Ansi)> _
    Structure PSN4_FixedHeader
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=8)> Dim headerID As Char()
        Dim varHdrLength As Int32
        Dim startTime As PSN4_DateTime
        Dim startTimeOffset As Double
        Dim sampleRate As Double
        Dim sampleCount As Int32
        Dim flags As Int32
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=3)> Dim timeRefType As Char()
        Dim timeRefStatus As Char
        Dim sampleType As Byte
        Dim sampleCompression As Byte
        Dim compIncident As Double
        Dim compAzimuth As Double
        Dim compOrientation As Char
        Dim sensorType As Byte
        Dim latitude As Double
        Dim longitude As Double
        Dim elevation As Double
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=6)> Dim name As Char()
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=4)> Dim compName As Char()
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=6)> Dim network As Char()
        Dim sensitivity As Double
        Dim magCorr As Double
        Dim atodBits As Int16
        Dim minimum As Double
        Dim maximum As Double
        Dim mean As Double
    End Structure


    ' ----------------------------------------------------- variable header
    <StructLayout(LayoutKind.Sequential, Pack:=1)> _
    Structure PSN4_VarHeader
        Dim checkNumber As Byte
        Dim id As DescID
        Dim length As Int32
        ' -------------------------------------------------- array of HeaderBytes (used for file read-write and CRC)
        Dim HeaderBytes As Byte()
        ' -------------------------------------------------- union of strings and structures
        Dim SensorLocString As String                     ' 1
        Dim SensorInfoString As String                    ' 2
        Dim CommentString As String                       ' 3
        Dim EventInfoStruct As PSN4_EventInfo             ' 4
        Dim PhasePickStruct As PSN4_PhasePick             ' 5
        Dim SampFilteringString As String                 ' 6
        Dim DataloggerIdString As String                  ' 7
        Dim IntSensorLocString As PSN4_InternationalText  ' 8
        Dim IntSensorInfoString As PSN4_InternationalText ' 9
        Dim IntCommentString As PSN4_InternationalText    ' 10
        Dim SensorADStruct As PSN4_SensorAD               ' 11
        Dim PolesAndZerosStruct As PSN4_PolesAndZeros     ' 12
        ' -------------------------------------------------- array of FieldBytes (used for file read-write and CRC)
        Dim FieldBytes As Byte()
    End Structure

    ' ----------------------------------------------------- variable length header - field descriptor ID numbers
    Enum DescID As Byte
        EndRecordMark = 0
        SensorLocString = 1      ' String - null terminated
        SensorInfoString = 2     ' String - null terminated  - e.g. Lehman, 4.5 hz geophone etc 
        CommentString = 3        ' String - null terminated
        EventInfoStruct = 4      ' EventInfo structure
        PhasePickStruct = 5      ' PhasePick structure
        SampFilteringString = 6  ' String - null terminated
        DataloggerIdString = 7   ' String - null terminated
        IntSensorLocString = 8   ' International Text structure  - UNUSED
        IntSensorInfoString = 9  ' International Text structure  - UNUSED
        IntCommentString = 10    ' International Text structure  - UNUSED
        SensorADStruct = 11      ' SensorAmpToD Structure ( SensorOutVoltage, AmpGain and A/D InputVolt )  - UNUSED
        PolesAndZerosStruct = 12 ' PolesAndZeros structure  - UNUSED
    End Enum

    ' ----------------------------------------------------- date and time ( 12 bytes )
    <StructLayout(LayoutKind.Sequential, Pack:=1)> _
    Structure PSN4_DateTime
        Dim year As UInt16
        Dim month As Byte
        Dim day As Byte
        Dim hour As Byte
        Dim minute As Byte
        Dim second As Byte
        Dim notUsed As Byte
        Dim nanoseconds As Int32
    End Structure

    '  ---------------------------------------------------- event informations ( 62 bytes )
    <StructLayout _
    (LayoutKind.Sequential, Pack:=1, CharSet:=CharSet.Ansi)> _
    Structure PSN4_EventInfo
        Dim eventTime As PSN4_DateTime
        Dim latitude As Double
        Dim longitude As Double
        Dim depth As Double
        Dim ms As Int16
        Dim mb As Int16
        Dim mw As Int16
        Dim ml As Int16
        Dim md As Int16
        Dim mOther As Int16
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=4)> Dim otherType As Char()
        Dim typeCode As Byte
        Dim qualityChar As Byte
        Dim flags As UInt16
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=6)> Dim agency As Char()
    End Structure

    ' ----------------------------------------------------- phase pick informations ( 42 bytes )
    <StructLayout _
    (LayoutKind.Sequential, Pack:=1, CharSet:=CharSet.Ansi)> _
    Structure PSN4_PhasePick
        Dim pickTime As PSN4_DateTime
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=8)> Dim phaseName As Char()
        Dim flags As UInt16
        Dim yLocation As Int16
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=16)> Dim fileName As Char()
        Dim tableDepth As Int16
    End Structure

    ' ----------------------------------------------------- International Text ( variable )
    <StructLayout _
    (LayoutKind.Sequential, Pack:=1, CharSet:=CharSet.Ansi)> _
    Structure PSN4_InternationalText
        <MarshalAs(UnmanagedType.ByValArray, SizeConst:=17)> Dim LanguageID As Char()
        <MarshalAs(UnmanagedType.AnsiBStr)> Dim text As String
    End Structure

    ' ----------------------------------------------------- SensorAmpAtoD ( 24 bytes )
    <StructLayout(LayoutKind.Sequential, Pack:=1)> _
    Structure PSN4_SensorAD
        Dim SensorOutputVoltage As Double
        Dim AmplifierGain As Double
        Dim AdInputVoltage As Double
    End Structure

    ' ----------------------------------------------------- PolesAndZeros ( variable )
    <StructLayout(LayoutKind.Sequential, Pack:=1)> _
    Structure PSN4_PolesAndZeros
        Dim NumberOfZeros As UInt16
        Dim NumberOfPoles As UInt16
        Dim Poles() As ComplexNumber
        Dim Zeros() As ComplexNumber
    End Structure

    Structure ComplexNumber
        Dim Real As Double
        Dim Imaginary As Double
    End Structure

    Private Sub InitPsnStructures(ByRef Psn As PSN4_File)
        Psn.Name = ""
        Psn.VolumeHeader.volumeIdAndVersion = "PSNVOLUME1".ToCharArray
        Psn.VolumeHeader.NumberEventRecords = 0
        ReDim Psn.Records(-1)
    End Sub



    ' ==========================================================================================
    '  LOAD and SAVE PsnFiles
    ' ==========================================================================================

    Dim PsnFile As PSN4_File

    Dim DataPath As String = Application.StartupPath & "\DataFiles\"

    Friend Sub SavePSN(ByVal fname As String)

        If Not fname.EndsWith(".psn", StringComparison.InvariantCultureIgnoreCase) Then
            fname &= ".psn"
        End If

        NewPSN(fname)
        NewPSN_AddRecord()

        Save_PsnFile(fname)

        PSN_Load(fname)
    End Sub

    Friend Sub NewPSN(ByVal fname As String)
        InitPsnStructures(PsnFile)
        PsnFile.Name = fname
    End Sub

    Friend Sub NewPSN_AddRecord()
        Dim i As Int32 = PsnFile.Records.Length
        PsnFile.VolumeHeader.NumberEventRecords = CUShort(i + 1)
        ReDim PsnFile.Records(i)
        ' ------------------------------------------------------------- FIXED HEADER
        With PsnFile.Records(i).FixedHeader
            .headerID = "PSNTYPE4".ToCharArray

            .varHdrLength = 6

            .startTime.year = 2012
            .startTime.month = 12
            .startTime.day = 12
            .startTime.hour = 12
            .startTime.minute = 12
            .startTime.second = 12
            .startTime.notUsed = 0
            .startTime.nanoseconds = 12

            .startTimeOffset = 0
            .sampleRate = 5
            .sampleCount = 10001
            .flags = 0
            .timeRefType = "WIN".ToCharArray()
            .timeRefStatus = CChar("L")
            .sampleType = 2
            .sampleCompression = 0
            .compIncident = 0
            .compAzimuth = 0
            .compOrientation = CChar("N")
            .sensorType = 0
            .latitude = 38.0
            .longitude = -87.0
            .elevation = 320
            .name = "      ".ToCharArray
            .compName = "    ".ToCharArray()
            .network = "      ".ToCharArray()
            .sensitivity = 0
            .magCorr = 0
            .atodBits = 16
            .minimum = 0
            .maximum = 0
            .mean = 0
        End With

        With PsnFile.Records(i)
            ' ------------------------------------------------------------- prepare FixedHeaderBytes for CRC
            .FixedHeaderBytes = StructureToByteArray(.FixedHeader)
            ' ------------------------------------------------------------- ONE EMPTY VAR HEADER
            ReDim .VarHeaders(0)
            .VarHeaders(0).checkNumber = 85
            .VarHeaders(0).id = 0
            .VarHeaders(0).length = 0
            ' ------------------------------------------------------------- DATA BYTES
            .NumDataBytes = 0
            ReDim .DataInt16(-1)
            ReDim .DataInt32(-1)
            ReDim .DataSingle(10000)
            ReDim .DataDouble(-1)

            .CRC = CalcCrcFromRecord(PsnFile.Records(i))
        End With


    End Sub



    ' ==========================================================================================
    '  LOAD PsnFile
    ' ==========================================================================================

    Friend Sub PSN_Load(ByVal fname As String)
        If Not fname.EndsWith(".psn", StringComparison.InvariantCultureIgnoreCase) Then
            fname &= ".psn"
        End If


        'Dim sw1 As Diagnostics.Stopwatch = New Diagnostics.Stopwatch
        'sw1.Reset()
        'sw1.Start()
        '
        Dim ReportString As String = ""



        ' -------------------------------------------------------------------------- init psn
        InitPsnStructures(PsnFile)
        PsnFile.Name = fname
        AddField(ReportString, "File", PsnFile.Name)
        ' -------------------------------------------------------------------------- open FileStream and BinaryReader
        Try
            Dim fStream As New FileStream(DataPath & fname, FileMode.Open, _
                                                            FileAccess.Read, _
                                                            FileShare.Read)
            Dim br As New BinaryReader(fStream)
            ' -------------------------------------------------------------------------- test PSNVOLUME1 and PSNTYPE4
            If fStream.Length > 10 Then
                Dim b As Byte() = br.ReadBytes(10)
                Dim s As String = System.Text.Encoding.ASCII.GetString(b)
                If s = "PSNVOLUME1" Then
                    PsnFile.VolumeHeader.NumberEventRecords = br.ReadUInt16
                    fStream.Position = 12
                Else
                    If Left(s, 8) = "PSNTYPE4" Then
                        PsnFile.VolumeHeader.NumberEventRecords = 1
                        fStream.Position = 0
                    Else
                        ReportString &= "Not a PSN Type4 file" & vbCrLf
                    End If
                End If
            End If
 

            ' -------------------------------------------------------------------------- unrecognized file
            If PsnFile.VolumeHeader.NumberEventRecords = 0 Then
                ReportString &= "Unrecognized file"
            End If

            ' -------------------------------------------------------------------------- read 1 or more records
            ReDim PsnFile.Records(PsnFile.VolumeHeader.NumberEventRecords - 1)
            For i As Int32 = 0 To PsnFile.VolumeHeader.NumberEventRecords - 1
                Load_PsnRecord(br, PsnFile.Records(i))
                ReportString &= Get_FixedHeader_String(PsnFile.Records(i).FixedHeader)
                ReportString &= Get_VarHeaders_String(PsnFile.Records(i).VarHeaders)
                ' ---------------------------------------------------------------------- CRC
                Dim CalcCRC As UInt16 = CalcCrcFromRecord(PsnFile.Records(i))
                AddField(ReportString, "------------------------", "")
                AddField(ReportString, "FileCRC", PsnFile.Records(i).CRC.ToString)
                AddField(ReportString, "CalcCRC", CalcCRC.ToString)
            Next
            ' -------------------------------------------------------------------------- close FileStream and BinaryReader
            br.Close()
            fStream.Close()

        Catch
            ReportString &= "Unrecognized file"
        End Try
        ' -------------------------------------------------------------------------- reports
        Form1.txt_Report.Text = ReportString
        Form1.txt_Report.SelectionStart = 1
        Form1.txt_Report.SelectionLength = 0

        'Form1.Text = (sw1.Elapsed.TotalMilliseconds * 1000).ToString("0")
        ' 18000 uS
    End Sub

    Private Sub Load_PsnRecord(ByVal br As BinaryReader, ByRef Record As PSN4_Record)
        ' --------------------------------------------------------------------------- read Fixed Header
        Read_FixedHeader(br, Record)
        ' --------------------------------------------------------------------------- read VarHeader
        ReDim Record.VarHeaders(-1)
        Do
            Dim n As Int32 = Record.VarHeaders.Length
            ReDim Preserve Record.VarHeaders(n)
            Read_VarHeader(br, Record.VarHeaders(n))
            If Record.VarHeaders(n).id = 0 Then Exit Do
            Read_VarHeaderData(br, Record.VarHeaders(n))
        Loop
        ' --------------------------------------------------------------------------- read samples
        Read_Samples(br, Record)
        ' --------------------------------------------------------------------------- read CRC
        Record.CRC = br.ReadUInt16
    End Sub

    Private Sub Read_FixedHeader(ByVal br As BinaryReader, ByRef record As PSN4_Record)
        ' ----------------------------------------------------------------- read header to ByteArray
        record.FixedHeaderBytes = br.ReadBytes(Marshal.SizeOf(record.FixedHeader))
        ' ----------------------------------------------------------------- copy ByteArray to structure FixedHeader
        record.FixedHeader = ByteArrayToStructure(Of PSN4_FixedHeader)(record.FixedHeaderBytes)
    End Sub

    Private Sub Read_VarHeader(ByVal br As BinaryReader, ByRef VarHeader As PSN4_VarHeader)
        VarHeader.checkNumber = br.ReadByte()
        VarHeader.id = CType(br.ReadByte(), DescID)
        VarHeader.length = br.ReadInt32
    End Sub

    Private Sub Read_VarHeaderData(ByVal br As BinaryReader, ByRef VarHeader As PSN4_VarHeader)
        VarHeader.FieldBytes = br.ReadBytes(VarHeader.length)
        Select Case VarHeader.id

            Case DescID.SensorLocString     ' string null terminated
                VarHeader.SensorLocString = Encoding.ASCII.GetString(VarHeader.FieldBytes, 0, VarHeader.length - 1)

            Case DescID.SensorInfoString    ' string null terminated  ( e.g. Lehman, 4.5 hz geophone etc )
                VarHeader.SensorInfoString = Encoding.ASCII.GetString(VarHeader.FieldBytes, 0, VarHeader.length - 1)

            Case DescID.CommentString       ' string null terminated
                VarHeader.CommentString = Encoding.ASCII.GetString(VarHeader.FieldBytes, 0, VarHeader.length - 1)

            Case DescID.EventInfoStruct     ' EventInfoStruct
                VarHeader.EventInfoStruct = ByteArrayToStructure(Of PSN4_EventInfo)(VarHeader.FieldBytes)

            Case DescID.PhasePickStruct     ' PhasePickStruct - UNUSED
                VarHeader.PhasePickStruct = ByteArrayToStructure(Of PSN4_PhasePick)(VarHeader.FieldBytes)

            Case DescID.SampFilteringString ' string null terminated 
                VarHeader.SampFilteringString = Encoding.ASCII.GetString(VarHeader.FieldBytes, 0, VarHeader.length - 1)

            Case DescID.DataloggerIdString  ' string null terminated
                VarHeader.DataloggerIdString = Encoding.ASCII.GetString(VarHeader.FieldBytes, 0, VarHeader.length - 1)

            Case DescID.IntSensorLocString  ' International Text structure  - UNUSED
                VarHeader.IntSensorLocString = ByteArrayToStructure(Of PSN4_InternationalText)(VarHeader.FieldBytes)

            Case DescID.IntSensorInfoString ' International Text structure  - UNUSED
                VarHeader.IntSensorInfoString = ByteArrayToStructure(Of PSN4_InternationalText)(VarHeader.FieldBytes)

            Case DescID.IntCommentString    ' International Text structure  - UNUSED
                VarHeader.IntCommentString = ByteArrayToStructure(Of PSN4_InternationalText)(VarHeader.FieldBytes)

            Case DescID.SensorADStruct      ' SensorAmpToD Structure - UNUSED
                VarHeader.SensorADStruct = ByteArrayToStructure(Of PSN4_SensorAD)(VarHeader.FieldBytes)

            Case DescID.PolesAndZerosStruct ' PolesAndZeros Structure  - UNUSED
                VarHeader.PolesAndZerosStruct = ByteArrayToStructure(Of PSN4_PolesAndZeros)(VarHeader.FieldBytes)

        End Select
    End Sub

    Private Sub Read_Samples(ByVal br As BinaryReader, ByRef record As PSN4_Record)
        Dim SamplesCount As Int32 = record.FixedHeader.sampleCount
        Select Case record.FixedHeader.sampleType
            Case 0
                record.NumDataBytes = SamplesCount * 2
                ReDim record.DataInt16(SamplesCount - 1)
                For i As Int32 = 0 To SamplesCount - 1
                    record.DataInt16(i) = br.ReadInt16()
                Next
            Case 1
                record.NumDataBytes = SamplesCount * 4
                ReDim record.DataInt32(SamplesCount - 1)
                For i As Int32 = 0 To SamplesCount - 1
                    record.DataInt32(i) = br.ReadInt32()
                Next
            Case 2
                record.NumDataBytes = SamplesCount * 4
                ReDim record.DataSingle(SamplesCount - 1)
                For i As Int32 = 0 To SamplesCount - 1
                    record.DataSingle(i) = br.ReadSingle()
                Next
            Case 3
                record.NumDataBytes = SamplesCount * 8
                ReDim record.DataDouble(SamplesCount - 1)
                For i As Int32 = 0 To SamplesCount - 1
                    record.DataDouble(i) = br.ReadDouble()
                Next
        End Select
    End Sub




    ' ==========================================================================================
    '  SAVE PsnFile
    ' ==========================================================================================

    Private Sub Save_PsnFile(ByVal fname As String)
        ' ----------------------------------------------------------------- test if the PsnFile is ready
        If PsnFile.VolumeHeader.NumberEventRecords = 0 Then Return
        ' ----------------------------------------------------------------- open FileStream and BinaryWriter
        Dim fStream As New FileStream(DataPath & fname, FileMode.Create, _
                                                        FileAccess.Write, _
                                                        FileShare.Write)
        Dim bw As New BinaryWriter(fStream)
        ' ----------------------------------------------------------------- write VOLUME HEADER
        If PsnFile.VolumeHeader.NumberEventRecords > 1 Then
            bw.Write(PsnFile.VolumeHeader.volumeIdAndVersion)
            bw.Write(PsnFile.VolumeHeader.NumberEventRecords)
        End If
        ' ----------------------------------------------------------------- write records
        For i As Int32 = 0 To PsnFile.VolumeHeader.NumberEventRecords - 1
            Save_PsnRecord(bw, PsnFile.Records(i))
        Next
        ' ----------------------------------------------------------------- close FileStream and BinaryWriter
        bw.Close()
        fStream.Close()
    End Sub

    Private Sub Save_PsnRecord(ByVal bw As BinaryWriter, ByRef Record As PSN4_Record)
        ' --------------------------------------------------------------------------- write Fixed Header
        Write_FixedHeader(bw, Record)
        ' --------------------------------------------------------------------------- write VarHeader
        Dim n As Int32 = 0
        Do
            Write_VarHeader(bw, Record.VarHeaders(n))
            If Record.VarHeaders(n).id = 0 Then Exit Do
            Write_VarHeaderData(bw, Record.VarHeaders(n))
            n += 1
        Loop
        ' --------------------------------------------------------------------------- write samples
        Write_Samples(bw, Record)
        ' --------------------------------------------------------------------------- write CRC
        If (Record.FixedHeader.flags And NO_CRC) = 0 Then
            bw.Write(Record.CRC)
        End If
    End Sub

    Private Sub Write_FixedHeader(ByVal bw As BinaryWriter, ByRef record As PSN4_Record)
        ' ----------------------------------------------------------------- copy structure FixedHeader to ByteArray
        'FixedHeaderStruct_To_ByteArray(record.FixedHeader, record.FixedHeaderBytes)
        record.FixedHeaderBytes = StructureToByteArray(record.FixedHeader)
        ' ----------------------------------------------------------------- write ByteArray to file
        bw.Write(record.FixedHeaderBytes)
    End Sub

    Private Sub Write_VarHeader(ByVal bw As BinaryWriter, ByRef VarHeader As PSN4_VarHeader)
        bw.Write(VarHeader.checkNumber)
        bw.Write(VarHeader.id)
        bw.Write(VarHeader.length)
    End Sub

    Private Sub Write_VarHeaderData(ByVal bw As BinaryWriter, ByRef VarHeader As PSN4_VarHeader)
        Select Case VarHeader.id

            Case DescID.SensorLocString     ' string null terminated
                VarHeader.FieldBytes = Encoding.ASCII.GetBytes(VarHeader.SensorLocString & Chr(0))

            Case DescID.SensorInfoString    ' string null terminated  ( e.g. Lehman, 4.5 hz geophone etc )
                VarHeader.FieldBytes = Encoding.ASCII.GetBytes(VarHeader.SensorInfoString & Chr(0))

            Case DescID.CommentString       ' string null terminated
                VarHeader.FieldBytes = Encoding.ASCII.GetBytes(VarHeader.CommentString & Chr(0))

            Case DescID.EventInfoStruct     ' EventInfo Structure
                VarHeader.FieldBytes = StructureToByteArray(VarHeader.EventInfoStruct)

            Case DescID.PhasePickStruct     ' PhasePick Structure - UNUSED
                VarHeader.FieldBytes = StructureToByteArray(VarHeader.PhasePickStruct)

            Case DescID.SampFilteringString ' string null terminated
                VarHeader.FieldBytes = Encoding.ASCII.GetBytes(VarHeader.SampFilteringString & Chr(0))

            Case DescID.DataloggerIdString  ' string null terminated
                VarHeader.FieldBytes = Encoding.ASCII.GetBytes(VarHeader.DataloggerIdString & Chr(0))

            Case DescID.IntSensorLocString  ' International Text structure  - UNUSED
                VarHeader.FieldBytes = StructureToByteArray(VarHeader.IntSensorLocString)

            Case DescID.IntSensorInfoString ' International Text structure  - UNUSED
                VarHeader.FieldBytes = StructureToByteArray(VarHeader.IntSensorInfoString)

            Case DescID.IntCommentString    ' International Text structure  - UNUSED
                VarHeader.FieldBytes = StructureToByteArray(VarHeader.IntCommentString)

            Case DescID.SensorADStruct      ' SensorAD Structure - UNUSED
                VarHeader.FieldBytes = StructureToByteArray(VarHeader.SensorADStruct)

            Case DescID.PolesAndZerosStruct ' PolesAndZeros Structure  - UNUSED
                VarHeader.FieldBytes = StructureToByteArray(VarHeader.PolesAndZerosStruct)

        End Select
        bw.Write(VarHeader.FieldBytes)
    End Sub

    Private Sub Write_Samples(ByVal bw As BinaryWriter, ByRef record As PSN4_Record)
        Select Case record.FixedHeader.sampleType
            Case 0
                For i As Int32 = 0 To record.DataInt16.Length - 1
                    bw.Write(record.DataInt16(i))
                Next
            Case 1
                For i As Int32 = 0 To record.DataInt32.Length - 1
                    bw.Write(record.DataInt32(i))
                Next
            Case 2
                For i As Int32 = 0 To record.DataSingle.Length - 1
                    bw.Write(record.DataSingle(i))
                Next
            Case 3
                For i As Int32 = 0 To record.DataDouble.Length - 1
                    bw.Write(record.DataDouble(i))
                Next
        End Select
    End Sub


    ' ==========================================================================================
    '  GENERIC - ByteArrays to Struct - NO CAST IS NEEDED
    ' ------------------------------------------------------------------------------------------
    '  calling method >>> FixedHeader = ByteArrayToStructure(Of PSN4_FixedHeader)(byteArray)
    ' ==========================================================================================
    Private Function ByteArrayToStructure(Of T)(ByRef bytearray() As Byte) As T
        Dim handle As GCHandle = GCHandle.Alloc(bytearray, GCHandleType.Pinned)
        ByteArrayToStructure = CType(Marshal.PtrToStructure(handle.AddrOfPinnedObject, GetType(T)), T)
        handle.Free()
    End Function

    ' ==========================================================================================
    '  GENERIC - Struct to ByteArray
    ' ==========================================================================================
    Private Function StructureToByteArray(Of T)(ByVal struct As T) As Byte()
        ReDim StructureToByteArray(Marshal.SizeOf(struct) - 1)
        Dim handle As GCHandle = GCHandle.Alloc(StructureToByteArray, GCHandleType.Pinned)
        Marshal.StructureToPtr(struct, handle.AddrOfPinnedObject, False)
        handle.Free()
    End Function



    ' ==========================================================================================
    '  CalcCrcFromRecord
    ' ==========================================================================================

    Private Function CalcCrcFromRecord(ByVal record As PSN4_Record) As UInt16
        CRC16_Init()
        ' ----------------------------------------------------------------- add fixed header
        CRC16_Add_ByteArray(record.FixedHeaderBytes)
        ' ----------------------------------------------------------------- add var headers
        For Each varheader As PSN4_VarHeader In record.VarHeaders
            CRC16_Add_Byte(varheader.checkNumber)
            CRC16_Add_Byte(varheader.id)
            CRC16_Add_Int32(varheader.length)
            If varheader.id > 0 Then
                CRC16_Add_ByteArray(varheader.FieldBytes)
            End If
        Next
        ' ----------------------------------------------------------------- add data
        Select Case record.FixedHeader.sampleType
            Case 0
                CRC16_Add_Int16_Array(record.DataInt16)
            Case 1
                CRC16_Add_Int32_Array(record.DataInt32)
            Case 2
                CRC16_Add_Single_Array(record.DataSingle)
            Case 3
                CRC16_Add_Double_Array(record.DataDouble)
        End Select
        ' ----------------------------------------------------------------- get CRC value
        CalcCrcFromRecord = CRC16_GetCrc()
    End Function



    ' ==========================================================================================
    '  REPORT STRINGS
    ' ==========================================================================================
    Private Function Get_FixedHeader_String(ByVal FixedHeader As PSN4_FixedHeader) As String
        Dim s As String = ""
        With FixedHeader
            AddField(s, "=========================", "")
            AddField(s, "    HeaderID", .headerID)
            AddField(s, "   VarHdrLen", .varHdrLength.ToString)
            AddDateTimeFields(s, .startTime)
            AddField(s, "StartTimeOfs", .startTimeOffset.ToString)
            AddField(s, "  SampleRate", .sampleRate.ToString("0.#########"))
            AddField(s, " SampleCount", .sampleCount.ToString)
            AddField(s, "       Flags", .flags.ToString)
            AddField(s, " TimeRefType", .timeRefType)
            AddField(s, "TimRefStatus", .timeRefStatus)
            AddField(s, "  SampleType", .sampleType.ToString)
            AddField(s, "  SampleComp", .sampleCompression.ToString)
            AddField(s, "CompIncident", .compIncident.ToString)
            AddField(s, " CompAzimuth", .compAzimuth.ToString("0.#########"))
            AddField(s, "  CompOrient", .compOrientation)
            AddField(s, "  SensorType", .sensorType.ToString)
            AddField(s, "    Latitude", .latitude.ToString("0.#########"))
            AddField(s, "   Longitude", .longitude.ToString("0.#########"))
            AddField(s, "   Elevation", .elevation.ToString("0.#########"))
            AddField(s, "        Name", .name)
            AddField(s, "    CompName", .compName)
            AddField(s, "     Network", .network)
            AddField(s, " Sensitivity", .sensitivity.ToString())
            AddField(s, "  MagnitCorr", .magCorr.ToString())
            AddField(s, "     AD_Bits", .atodBits.ToString)
            AddField(s, "     Minimum", .minimum.ToString("0.#########"))
            AddField(s, "     Maximum", .maximum.ToString("0.#########"))
            AddField(s, "        Mean", .mean.ToString("0.#########"))
        End With
        Return s
    End Function

    Private Sub AddDateTimeFields(ByRef s As String, ByVal datetime As PSN4_DateTime)
        AddField(s, "   StartDate", datetime.year.ToString & "/" & _
                                    datetime.month.ToString & "/" & _
                                    datetime.day.ToString)
        AddField(s, "   StartTime", datetime.hour.ToString & ":" & _
                                    datetime.minute.ToString & ":" & _
                                    datetime.second.ToString)
        AddField(s, "StartNanoSec", datetime.nanoseconds.ToString)
    End Sub


    Private Function Get_VarHeaders_String(ByVal varHeaders As PSN4_VarHeader()) As String
        Dim s As String = ""
        AddField(s, "-------------------------", "")
        For Each vh As PSN4_VarHeader In varHeaders
            Select Case vh.id
                Case DescID.SensorLocString
                    AddField(s, "SensorLoc ---------------" & vbCrLf, vh.SensorLocString)
                Case DescID.SensorInfoString
                    AddField(s, "SensorInfo --------------" & vbCrLf, vh.SensorInfoString)
                Case DescID.CommentString
                    AddField(s, "Comment -----------------" & vbCrLf, Replace(vh.CommentString, ",", vbCrLf & " "))
                Case DescID.EventInfoStruct
                    AddField(s, "EventInfo ---------------", "")
                    AddDateTimeFields(s, vh.EventInfoStruct.eventTime)
                    AddField(s, "    Latitude", vh.EventInfoStruct.latitude.ToString("0.#########"))
                    AddField(s, "   Longitude", vh.EventInfoStruct.longitude.ToString("0.#########"))
                    AddField(s, "       Depth", vh.EventInfoStruct.depth.ToString("0.######"))
                    AddField(s, "Magnitude Ms", (vh.EventInfoStruct.ms / 100).ToString("0.######"))
                    AddField(s, "Magnitude Mb", (vh.EventInfoStruct.mb / 100).ToString("0.######"))
                    AddField(s, "Magnitude Mw", (vh.EventInfoStruct.mw / 100).ToString("0.######"))
                    AddField(s, "Magnitude Ml", (vh.EventInfoStruct.ml / 100).ToString("0.######"))
                    AddField(s, "Magnitude Md", (vh.EventInfoStruct.md / 100).ToString("0.######"))
                    AddField(s, " Magn. Other", (vh.EventInfoStruct.mOther / 100).ToString("0.######"))
                    AddField(s, "MgnOtherType", vh.EventInfoStruct.otherType)
                    AddField(s, "     Quality", (vh.EventInfoStruct.qualityChar).ToString())
                    AddField(s, "       Flags", (vh.EventInfoStruct.flags).ToString())
                    AddField(s, "      Agency", vh.EventInfoStruct.agency)
                Case DescID.PhasePickStruct
                    AddField(s, "PhasePick ---------------", "UNIMPLEMENTED")
                Case DescID.SampFilteringString
                    AddField(s, "SampFiltering -----------" & vbCrLf, Replace(vh.SampFilteringString, ",", vbCrLf & " "))
                Case DescID.DataloggerIdString
                    AddField(s, "DataloggerId  -----------" & vbCrLf, vh.DataloggerIdString)
                Case DescID.IntSensorLocString
                    AddField(s, "Int. Locality -----------", "UNIMPLEMENTED")
                Case DescID.IntSensorInfoString
                    AddField(s, "Int. SensorInfo ---------", "UNIMPLEMENTED")
                Case DescID.IntCommentString
                    AddField(s, "Int. Comment ------------", "UNIMPLEMENTED")
                Case DescID.SensorADStruct
                    AddField(s, "Sensor AD ---------------", "UNIMPLEMENTED")
                Case DescID.PolesAndZerosStruct
                    AddField(s, "Poles and Zeros ---------", "UNIMPLEMENTED")
            End Select
        Next
        Return s
    End Function

    Private Sub AddField(ByRef s As String, ByVal name As String, ByVal value As String)
        s &= name & " " & value.Replace(Chr(0), Chr(32)) & vbCrLf
    End Sub

End Module
